/*

Devanagari codes are in this order:

a,aa,i,ii,u,uu,
.r,.r.r,.l,.l.l,
e,ai,o,au,
k,kh,g,gh,"n,
c,ch,k,jh,~n,
.t, .th, .d, .dh, .n,
t,th,d,dh,n,
p,ph,b,bh,m,
y,r,l,v,
"s,.s,s,h,
.h,.m


*/

enum sa {
	SA_A,SA_AA,SA_I,SA_II,SA_U,SA_UU,
	SA_R_VOWEL,SA_RR_VOWEL,SA_L_VOWEL,SA_LL_VOWEL,
	SA_E,SA_AI,SA_O,SA_AU,
	SA_K,SA_KH,SA_G,SA_GH,SA_N_VELAR,
	SA_C,SA_CH,SA_J,SA_JH,SA_N_PALATAL,
	SA_T_RETRO,SA_TH_RETRO,SA_D_RETRO,SA_DH_RETRO,SA_N_RETRO,
	SA_T,SA_TH,SA_D,SA_DH,SA_N,
	SA_P,SA_PH,SA_B,SA_BH,SA_M,
	SA_Y,SA_R,SA_L,SA_V,
	SA_S_PALATAL,SA_S_RETRO,SA_S,SA_H,
	SA_VISARGA,SA_ANUSVEDA,SA_MAX};

unsigned short dev[SA_MAX]=
{0x0905,0x0906,0x0907,0x0908,0x0909,0x090a,
0x090b,0x0960,0x090c,0x0961,
0x090f,0x0910,0x0913,0x0914,
0x0915,0x0916,0x0917,0x0918,0x0919,
0x091a,0x091b,0x091c,0x091d,0x091e,
0x091f,0x0920,0x0921,0x0922,0x0923,
0x0924,0x0925,0x0926,0x0927,0x0928,
0x092a,0x092b,0x092c,0x092d,0x092e,
0x092f,0x0930,0x0932,0x0935,
0x0936,0x0937,0x0938,0x0939,
0x0903,0x0902};

unsigned short cvowels[SA_K]=
{0,0x093e,0x093f,0x0940,0x0941,0x0942,
0x0943,0x0944,0x0962,0x0963,
0x0947,0x0948,0x094b,0x094c};

#define HALANTA 0x094d

unsigned short iast[SA_MAX]=
{'a',0x0101,'i',0x012b,'u',0x016b,
0x1e5b,0x1e5d,0x1e37,0x1e39,
'e','a','o','a',
'k','k','g','g',0x1e45,
'c','c','j','j',0x00f1,
0x1e6d,0x1e6d,0x1e0d,0x1e0d,0x1e47,
't','t','d','d','n',
'p','p','b','b','m',
'y','r','l','v',
0x015b,0x1E63,'s','h',
0x1e25,0x1e43};

const unsigned short iasts[SA_MAX][3]=
{{'a',0,0},{0x0101,0,0},{'i',0,0},{0x012b,0,0},{'u',0,0},{0x016b,0,0},
{0x1e5b,0,0},{0x1e5d,0,0},{0x1e37,0,0},{0x1e39,0,0},
{'e',0,0},{'a','i',0},{'o',0,0},{'a','u',0},
{'k',0,0},{'k','h',0},{'g',0,0},{'g','h',0},{0x1e45,0,0},
{'c',0,0},{'c','h',0},{'j',0,0},{'j','h',0},{0x00f1,0,0},
{0x1e6d,0,0},{0x1e6d,'h',0},{0x10d,0,0},{0x10d,'h',0},{0x1e47,0,0},
{'t',0,0},{'t','h',0},{'d',0,0},{'d','h',0},{'n',0,0},
{'p',0,0},{'p','h',0},{'b',0,0},{'b','h',0},{'m',0,0},
{'y',0,0},{'r',0,0},{'l',0,0},{'v',0,0},
{0x015b,0,0},{0x1e63,0,0},{'s',0,0},{'h',0,0},
{0x1e25,0,0},{0x1e43,0,0}};

const char* itrans[SA_MAX]=
{"a","aa","i","ii","u","uu",
".r",".r.r",".l",".l.l",
"e","ai","o","au",
"k","kh","g","gh","\"n",
"c","ch","j","jh","~n",
".t",".th",".d",".dh",".n",
"t","th","d","dh","n",
"p","ph","b","bh","m",
"y","r","l","v",
"\"s",".s","s","h",
".h",".m"};

unsigned short capforms[]=
{
	0x1E24,0x1E42,0x0100,0x012A,0x016A,0x1E5a,0x1E5c,
	0x1E36,0x1E38,0x1E44,0x1E6C,0x1E0C,0x1E46,0x015A,
	0x1E62,0x0112,0x014C
};

#define NCAPFORMS (sizeof(capforms)/sizeof(capforms[0]))

#define A_CIRC_UC 0x00c2
#define E_CIRC_UC 0x00ca
#define O_CIRC_UC 0x00d4
#define A_CIRC 0x00e2
#define E_CIRC 0x00ea
#define E_MACRON 0x0113
#define O_CIRC 0x00f4
#define O_MACRON 0x014d

bool unaspirated(sa s)
{
	return s==SA_K||s==SA_G||s==SA_C||s==SA_J||s==SA_T_RETRO||
		s==SA_D_RETRO||s==SA_T||s==SA_D||s==SA_P||s==SA_B;
}

sa transcode_char_2(unsigned short* s,int& k)
{
	int ib=-1; int kb=k;
	for (int i=0;i<SA_MAX;i++) {
		const char* key=itrans[i];
		int kk=k;
		bool b=true;
		while (*key) {
			if (*key++ != s[kk++]) { b=false; break; }
		}
		if (b && kk > kb) {
			kb=kk;
			ib=i;
		}
	}

	if (ib<0) {
		k++; return SA_MAX;
	}
	else  {
		k=kb; return (sa) ib;
	}
}		

bool capform(unsigned short x)
{
	for (int i=0;i<NCAPFORMS;i++) {
		if (x==capforms[i]) return true;
	}
	return false;
}

unsigned short cf(unsigned short x)
{
	if (x >= 'A' && x <= 'Z') 
		return x-'A'+'a';
	if (x==0x00D1) return 0x00F1;
	if (capform(x)) return x+1;
// get rid of sandhi initial accents	
	if (x==A_CIRC||x==A_CIRC_UC) x=0x0101;
	if (x==E_CIRC||x==E_CIRC_UC||x==E_MACRON) x='e';
	if (x==O_CIRC||x==O_CIRC_UC||x==O_MACRON) x='o';
	return x;
}

sa transcode_char(unsigned short* s,int& k)
{
	int j;
	for (j=0;j<SA_MAX;j++)
		if (cf(s[k])==iast[j]) break;
	k++;
	if (j==SA_MAX) return SA_MAX;
	if (unaspirated((sa) j) && s[k]=='h') {
		k++; j++;
	}
	if (j==SA_A && s[k]=='i')  {
		k++; j=SA_AI;
	}
	if (j==SA_A && s[k]=='u')  {
		k++; j=SA_AU;
	}
	return (sa) j;
}

bool consonant(sa s)
{
	return s >= SA_K && s <= SA_H;
}

bool vowel(sa s)
{
	return s <= SA_AU;
}

bool isblank(unsigned short x)
{
	return x==' '||x=='\''||x==0x201d||x==0x2019;
}

int sl(const unsigned short* c)
{
	int i=0;
	while (c[i]) i++;
	return i;
}

void sc(unsigned short* t,const unsigned short* s)
{
	int i=0;
	while (s[i]) t[i++]=s[i];
	t[i]=0;
	return;
}


unsigned short* transcode2(unsigned short* c)
{
	int i=0; int t=0;
	while (c[i]) {
		sa x=transcode_char_2(c,i);
		if (x==SA_MAX) t++;
		else t+=sl(iasts[x]);
	}
	unsigned short *w=new unsigned short [t+1];
	i=0; t=0; unsigned short cc;
	while (cc=c[i]) {
		sa x=transcode_char_2(c,i);
		if (x==SA_MAX) w[t++]=cc;
		else {
			sc(w+t,iasts[x]);
			t+=sl(iasts[x]);
		}
	}
	w[t]=0;
	return w;
}




unsigned short* transcode(unsigned short* c)
{
	int i=0; int t=0;
	bool bCons=false;
	while (c[i]) {
		if (isblank(c[i])) {
			i++; continue;
		}
		sa x=transcode_char(c,i);
		if (bCons && x==SA_A) ;
		else if (bCons && !vowel(x))
			t+=2;
		else t++;
		bCons=consonant(x);
	}
	if (bCons) t++;
	unsigned short *w=new unsigned short [t+1];
	i=0; t=0;
	unsigned short cc;
	bCons=false;
	while (cc=c[i]) {
		if (isblank(c[i])) {
			i++; continue;
		}
		sa x=transcode_char(c,i);
		if (bCons && x==SA_A);
		else if (bCons && !vowel(x)) {
			w[t++]=HALANTA; w[t++]=(x==SA_MAX?cc:dev[x]);
		}
		else if (bCons && vowel(x)) {
			w[t++]=cvowels[x];
		}
		else w[t++]=(x==SA_MAX?cc:dev[x]);
		bCons=consonant(x);
	}
	if (bCons) w[t++]=HALANTA;
	w[t]=0;
	return w;
}



