#ifdef _WIN32
#include "stdafx.h"
#include "stdio.h"
#include <windows.h>
#else
#include "stdio.h"
#include "string.h"
#endif

// This needs cleaning up so that in Windows too xaira-php lives in the
// xara tree not the php one
#ifdef _WIN32
#include "XairoHeader.h" 
#else
#include "../XairoHeader.h"
#endif
#include "XairoPHPError.h"

int kserver=0;

char* utf8(const xairo_uchar* w);
xairo_uchar* uc(const char* w);
extern "C" void xperr(const char*);

unsigned short* transcode(unsigned short* c);
unsigned short* transcode2(unsigned short* c);

using namespace XAIRO_NAMESPACE;


CXairoInterface* theServer=NULL;
CXairoInterface* theServer2=NULL;
CXairoInterface* theServer3=NULL;

#ifdef _WIN32
#define SEPCHAR '\\'
#else
#define SEPCHAR '/'
#ifndef DATAPATH
#error The symbol DATAPATH must be defined.
#endif
#define _strdup strdup
#endif

#define _MAX_PATH 128

void debuglog(const char* s)
{
	FILE* f=fopen("/home/xaira/dlog","a");
	fprintf(f,"%s\n",s);
	fclose(f);
}

class CPHPXairoBootMonitor : public CXairoIBootMonitor
{
public :
	CPHPXairoBootMonitor() {}
	void Log(const xairo_uchar*);
};

void CPHPXairoBootMonitor::Log(const xairo_uchar* s)
{
	xperr(utf8(s));
}

static CXairoIParameters xp;

struct _inicorpus {
	char name[32];
	char path[_MAX_PATH];
	CXairoInterface* server;
};

static int ncorp=0;

static _inicorpus *corpora=NULL;


const char* _getTempPath()
{
	return xp.temppath;
}

const char* _getDataPath()
{
	return xp.datapath;
}

bool parsecp(const char* cp)
{
	if (cp[0]==0) return true;
	const char* p0=cp;
	int k=0;
	while (p0) {
		k++;
		p0=strchr(p0,';');
		if (p0) p0++;
	}
	ncorp=k;
	corpora=new _inicorpus [k];
	k=0;
	do {
        const char* p=strchr(cp,';');
		int kp=p?(p-cp):strlen(cp);
		const char* e=strchr(cp,'=');
		int ke=e?(e-cp):strlen(cp);
		if (ke >= kp) {
			sprintf(corpora[k].name,"corpus%d",k+1);
			ke=-1;
		} else {
			strncpy(corpora[k].name,cp,ke);
			corpora[k].name[ke]=0;
		}
		strncpy(corpora[k].path,cp+ke+1,kp-ke-1);
		corpora[k].path[kp-ke-1]=0;
		k++;
		cp=p?(p+1):NULL;
	}
	while (cp);
	return true;
}



bool xairainit(const char* cp,const char* dp,const char* tp)
{
	if (!parsecp(cp)) {
		xperr("Error in corpuspath parameter");
		return false;
	}
	if (ncorp==0) return true;
	CPHPXairoBootMonitor b;
//	xp.parameter_file=cp;
	xp.bm=&b;
	xp.inth=NULL;
	static char xtp[_MAX_PATH];
	if (tp[0]) strcpy(xtp,tp);
	else
#ifdef _WIN32
		GetTempPathA(_MAX_PATH,xtp);
#else 
		strcpy(xtp,"/tmp");
#endif
	xp.temppath=xtp;
	static char xdp[_MAX_PATH];
	if (dp[0]) {
		strcpy(xdp,dp);
		if (xdp[strlen(xdp)-1]==SEPCHAR) xdp[strlen(xdp)-1]=0;
	} else {
#ifdef _WIN32
		GetModuleFileNameA(NULL,xdp,_MAX_PATH);
		char* x=strrchr(xdp,SEPCHAR);
		*x=0;
#else
		strcpy(xdp,DATAPATH);
#endif
	}
	xp.datapath=xdp;
	for (int i=0;i<ncorp;i++) {
		xp.parameter_file=corpora[i].path;
		CXairoInterface* t=CXairoImplementation::getServer(&xp);
		if (t==NULL) return false;
		corpora[i].server=t;
	}
	return true;
}

void* _getServer(const char* name) {
	if (name==NULL && ncorp > 0) 
		return corpora[0].server;
	for (int i=0;i<ncorp;i++) {
		if (strcmp(corpora[i].name,name)==0) 
			return corpora[i].server;
	}
	return NULL;
}

const char* _getCorpusName(void* v)
{
	return utf8(((CXairoInterface*) v)->getCorpusName());
}

const char* gcn(int i)
{
	return corpora[i].name;
}

const char* gcp(int i)
{
	return corpora[i].path;
}

const char* _getCorpusPath(void* v)
{
	for (int i=0;i<ncorp;i++) 
		if (v==corpora[i].server) return corpora[i].path;
	return "";	
}

int gcc()
{
	return ncorp;
}

void* _getGrammar(void * v) {
	return (void*) ((CXairoInterface*)v)->getGrammar();
// OK I've cast away he qualifier but DON'T delete it
}

void* _deserialize(void * v,const char* c)
{
	return ((CXairoInterface*) v)->deserialize(uc(c));
}

int Grammar_getAddkeyCount(void* v)
{
	return ((CXairoIGrammar*) v)->getAddkeyCount();
}

void* Grammar_getAddkey(void* v,int n)
{
	return (void*) ((CXairoIGrammar*) v)->getAddkey(n);
}

int Grammar_getLemmaSchemeCount(void* v)
{
	return ((CXairoIGrammar*) v)->getLemmaSchemeCount();
}

void* Grammar_getLemmaScheme(void* v,int n)
{
	try {
		return (void*) ((CXairoIGrammar*) v)->getLemmaScheme(n);
	}
	catch (CXairoIError* e)
	{
		throw new CXairoPHPError(e->getMessage());
	}
}

void* Grammar_getRootPartition(void* v)
{
	return (void*) ((CXairoIGrammar*) v)->getRootPartition();
}

int Grammar_getElementCount(void* v)
{
	return ((CXairoIGrammar*) v)->getElementCount();
}

void* Grammar_getElement(void* v,int n)
{
	try {
		return (void*) ((CXairoIGrammar*) v)->getElement(n);
	}
	catch (CXairoIError* e)
	{
		throw new CXairoPHPError(e->getMessage());
	}
}


char* GrammarAddkey_getName(void* v)
{
	return utf8(((CXairoIGrammar::Addkey*) v)->getName());
}

bool GrammarAddkey_isLemmaScheme(void* v)
{
	return ((CXairoIGrammar::Addkey*) v)->isLemmaScheme();
}

char* GrammarLemmaScheme_getName(void* v)
{
	return utf8(((CXairoIGrammar::LemmaScheme*) v)->getName());
}

char* GrammarLemmaScheme_getGloss(void* v)
{
	return utf8(((CXairoIGrammar::LemmaScheme*) v)->getGloss());
}

int GrammarLemmaScheme_getFacetCount(void* v)
{
	return ((CXairoIGrammar::LemmaScheme*) v)->getFacetCount();
}

char* GrammarLemmaScheme_getFacet(void* v,int i)
{
	return utf8(((CXairoIGrammar::LemmaScheme*) v)->getFacet(i));
}

char* GrammarPartition_getName(void* v)
{
	return utf8(((CXairoIGrammar::Partition*) v)->getName());
}

char* GrammarPartition_getDescription(void* v)
{
	return utf8(((CXairoIGrammar::Partition*) v)->getDescription());
}

int GrammarPartition_getCount(void* v)
{
	return ((CXairoIGrammar::Partition*) v)->getCount();
}

void* GrammarPartition_getDivision(void* v,int n)
{
	return (void*) ((CXairoIGrammar::Partition*) v)->getDivision(n);
}

char* GrammarElement_getName(void* v)
{
	return utf8(((CXairoIGrammar::Element*) v)->getName());
}

char* GrammarElement_getGloss(void* v)
{
	return utf8(((CXairoIGrammar::Element*) v)->getGloss());
}

char* GrammarElement_getNamespace(void* v)
{
	return utf8(((CXairoIGrammar::Element*) v)->getNamespace()->getUrl());
}

int GrammarElement_getAttributeCount(void* v)
{
	return ((CXairoIGrammar::Element*) v)->getAttributeCount();
}

void* GrammarElement_getAttribute(void* v,int n)
{
	try {
		return (void*) ((CXairoIGrammar::Element*) v)->getAttribute(n);
	}
	catch (CXairoIError* e)
	{
		throw new CXairoPHPError(e->getMessage());
	}
}


char* GrammarAttribute_getName(void* v)
{
	return utf8(((CXairoIGrammar::Attribute*) v)->getName());
}

char* GrammarAttribute_getGloss(void* v)
{
	return utf8(((CXairoIGrammar::Attribute*) v)->getGloss());
}

char* GrammarAttribute_getNamespace(void* v)
{
	return utf8(((CXairoIGrammar::Attribute*) v)->getNamespace()->getUrl());
}


void* _getWordList(void* v,const char* s,const char* ls)
{
	const xairo_uchar* ss=uc(s);
	try {
		CXairoIWordList* l=((CXairoInterface*)v)->getWordList(1,&ss,uc(ls));
		return l;
	}
	catch (CXairoIError* e)
	{
		throw new CXairoPHPError(e->getMessage());
	}
}

void* _getAddkeyValueList(void* v,void* ak)
{
	try {
		CXairoIAddkeyValueList* l=((CXairoInterface*)v)->getAddkeyValueList((CXairoIGrammar::Addkey*) ak);
		return l;
	}
	catch (CXairoIError* e)
	{
		throw new CXairoPHPError(e->getMessage());
	}
}

void* _getAttributeValueList(void* v,void* a)
{
	try {
		CXairoIAttributeValueList* l=((CXairoInterface*)v)->getAttributeValueList((CXairoIGrammar::Attribute*) a);
		return l;
	}
	catch (CXairoIError* e)
	{
		throw new CXairoPHPError(e->getMessage());
	}
}

void* _getFrequencyList(void* v,const char* s,int maxn,int minf,int maxf,const char* ls)
{
	const xairo_uchar* ss=uc(s);
	try {
		CXairoIWordList* l=((CXairoInterface*)v)->getFrequencyList(1,&ss,maxn,minf,maxf,uc(ls));
		return l;
	}
	catch (CXairoIError* e)
	{
		throw new CXairoPHPError(e->getMessage());
	}
}

void* _getFilteredFrequencyList(void* v,const char* s,int maxn,int minf,int maxf,const char* ls,void*vf)
{
	const xairo_uchar* ss=uc(s);
	try {
		CXairoIWordList* l=((CXairoInterface*)v)->getFilteredFrequencyList(1,&ss,maxn,minf,maxf,uc(ls),(CXairoICorpusTextFilter*) vf,NULL);
		return l;
	}
	catch (CXairoIError* e)
	{
		throw new CXairoPHPError(e->getMessage());
	}
}



void* _getUsage(void* v,int i)
{
	return ((CXairoInterface*) v)->getUsage(i);
}

void* _solve(void* v,const char* s)
{
	try {
		CXairoISolution* sl=((CXairoInterface*)v)->solve(uc(s));
		return sl;
	}
	catch (CXairoIError* e)
	{
		throw new CXairoPHPError(e->getMessage());
	}
}

void* _getCorpusTextList(void * v)
{
	return ((CXairoInterface*) v)->getCorpusTextList();
}

int WordList_getCount(void* v)
{
	return ((CXairoIWordList*) v)->getCount();
}

int AddkeyValueList_getCount(void* v)
{
	return ((CXairoIAddkeyValueList*) v)->getCount();
}

int AttributeValueList_getCount(void* v)
{
	return ((CXairoIAttributeValueList*) v)->getCount();
}

int Solution_getCount(void* v)
{
	return ((CXairoISolution*) v)->getCount();
}

char* Solution_serialize(void* v)
{
	return utf8(((CXairoISolution*) v) ->serialize());
}


void* Solution_thinCorpusTextFilter(void* v,void* vv)
{
	return ((CXairoISolution*) v)->thinCorpusTextFilter((CXairoICorpusTextFilter*) vv);
}

void* Solution_thinRand(void* v,int n,int s)
{
	return ((CXairoISolution*) v)->thinRand(n,s);	
}

void* Solution_thinInit(void* v,int n)
{
	return ((CXairoISolution*) v)->thinInit(n);	
}

void* Solution_thinOpt(void* v)
{
	return ((CXairoISolution*) v)->thinOnePerText();	
}

void* Solution_thinSelection(void* v,int n,int* p)
{
	return ((CXairoISolution*) v)->thinSelection(n,p);	
}

void* Solution_thinCollocation(void* v,const char* ls,int wl,int wr,int n,const char** cc)
{
	const xairo_uchar** cc2=new const xairo_uchar*[n];
	for (int i=0;i<n;i++) cc2[i]=uc(cc[i]);
	void *vv=((CXairoISolution*) v)->thinCollocation(uc(ls),CXairoICollocationTable::_cwin(wl,wr),n,cc2);
	delete [] cc2;
	return vv;
}

void* Solution_thinCollocationX(void* v,const char* ls,const char* el,const char* ns,int n,const char** cc)
{
	const xairo_uchar** cc2=new const xairo_uchar*[n];
	for (int i=0;i<n;i++) cc2[i]=uc(cc[i]);
	void *vv=((CXairoISolution*) v)->thinCollocation(uc(ls),CXairoICollocationTable::_cwin(uc(el),uc(ns)),n,cc2);
	delete [] cc2;
	return vv;
}

void Solution_analysis(void* v,int np,int* p,int nb,int*a,int*at)
{
	((CXairoISolution*) v)->getAnalysis(np,p,nb,a,at);
}

void* Solution_getHit(void* v,int k)
{
	return ((CXairoISolution*) v)->getHit(k);
}

void* Solution_makeCollocationTable(void* v,const char* p,const char* k,int m, int wl,int wr,
	int maxn,double mins,double maxs,int lowf)
{
	return ((CXairoISolution*) v)->makeCollocationTable(uc(p),uc(k),(CXairoICollocationTable::cmeth) m,CXairoICollocationTable::_cwin(wl,wr),maxn,mins,maxs,lowf);
}

void* Solution_makeCollocationTableX(void* v,const char* p,const char* k,int m, const char* el,const char* ns,
	int maxn,double mins,double maxs,int lowf)
{
	return ((CXairoISolution*) v)->makeCollocationTable(uc(p),uc(k),(CXairoICollocationTable::cmeth) m,CXairoICollocationTable::_cwin(uc(el),uc(ns)),maxn,mins,maxs,lowf);
}

void* WordList_getWord(void* v,int i)
{
	return ((CXairoIWordList*) v)->getWord(i);
}

void* AddkeyValueList_getAddkeyValue(void* v,int i)
{
	return ((CXairoIAddkeyValueList*) v)->getAddkeyValue(i);
}

void* AttributeValueList_getAttributeValue(void* v,int i)
{
	return ((CXairoIAttributeValueList*) v)->getAttributeValue(i);
}

int Word_getFrequency(void* v)
{
	return ((CXairoIWord*) v)->getFrequency();
}

int Word_getFilteredFrequency(void* v,void* vv)
{
	return ((CXairoIWord*) v)->getFilteredFrequency((CXairoICorpusTextFilter*) vv,NULL);
}

int Word_getFormCount(void* v)
{
	return ((CXairoIWord*) v)->getFormCount();
}

char* Word_getNorm(void* v,int n)
{
	return utf8(((CXairoIWord*) v)->getNorm(n));
}

void* Word_getForm(void* v,int n)
{
	return ((CXairoIWord*) v)->getForm(n);
}

char* AddkeyValue_getValue(void* v)
{
	return utf8(((CXairoIAddkeyValue*) v)->getValue());
}

int AddkeyValue_getFrequency(void* v)
{
	return ((CXairoIAddkeyValue*) v)->getFrequency();
}

char* AttributeValue_getValue(void* v)
{
	return utf8(((CXairoIAttributeValue*) v)->getValue());
}

int AttributeValue_getFrequency(void* v)
{
	return ((CXairoIAttributeValue*) v)->getFrequency();
}

int AttributeValue_getFilteredFrequency(void* v,void* vv)
{
	return ((CXairoIAttributeValue*) v)->getFilteredFrequency((CXairoICorpusTextFilter*) vv);
}


char* Form_getNorm(void* v)
{
	return utf8(((CXairoIForm*) v)->getNorm());
}

char* Form_getText(void* v)
{
	return utf8(((CXairoIForm*) v)->getText());
}

int Form_getFrequency(void* v)
{
	return ((CXairoIForm*) v)->getFrequency();
}

int Form_getFilteredFrequency(void* v,void* vv)
{
	return ((CXairoIForm*) v)->getFilteredFrequency((CXairoICorpusTextFilter*) vv);
}

char* Form_getAdditionalKey(void* v,const char* k)
{
	return utf8(((CXairoIForm*) v)->getAdditionalKey(uc(k)));
}

char* Hit_getText(void* v,int n,int m)
{
	return utf8(((CXairoIHit*) v)->getText(n,m));
}

void* Hit_getCorpusText(void* v)
{
	return ((CXairoIHit*) v)->getCorpusText();
}

char* Hit_getUnit(void* v)
{
	return utf8(((CXairoIHit*) v)->getUnit());
}

int Hit_getCursorStart(void* v)
{
	return ((CXairoIHit*) v)->getCursorStart();
}

int Hit_getCursorEnd(void* v)
{
	return ((CXairoIHit*) v)->getCursorEnd();
}

char* CorpusText_getTitle(void* v)
{
	return utf8(((CXairoICorpusText*) v)->getTitle());
}

void* CorpusText_evalXpath(void* v,const char* ex)
{
	return ((CXairoICorpusText*) v)->evalXpath(uc(ex));
}

bool CorpusText_isSelected(void* v,void* vv)
{
	return ((CXairoICorpusText*) v)->isSelected((CXairoICorpusTextFilter*) vv);
}

int CorpusText_classify(void* v,void* vv)
{
	return ((CXairoICorpusText*) v)->classify((CXairoIGrammar::Partition*) vv);
}

int CorpusText_getWordCount(void* v)
{
	return ((CXairoICorpusText*) v)->getWordCount();
}



char* XpathValue_getText(void* v)
{
	return utf8(((CXairoIXpathValue*) v)->getText());
}


int CorpusTextList_getCount(void* v)
{
	return ((CXairoICorpusTextList*) v)->getCount();
}

void* CorpusTextList_getCorpusText(void* v,int n)
{
	return ((CXairoICorpusTextList*) v)->getCorpusText(n);
}

void* CorpusTextList_getCorpusTextFilter(void* v,int nt,int* t)
{
	return ((CXairoICorpusTextList*) v)->getCorpusTextFilter(nt,t);
}

char* CorpusTextFilter_serialize(void* v)
{
	return utf8(((CXairoICorpusTextFilter*) v) ->serialize());
}

int CollocationTable_getCount(void* v)
{
	return ((CXairoICollocationTable*) v)->getCount();
}

void* CollocationTable_getCollocate(void* v,int n)
{
	return ((CXairoICollocationTable*) v)->getCollocate(n);
}

void* Collocate_getWord(void* v)
{
	return ((CXairoICollocate*) v)->getWord();
}



int Collocate_getHits(void* v)
{
	return ((CXairoICollocate*) v)->getHits();
}

double Collocate_getScore(void* v)
{
	return ((CXairoICollocate*) v)->getScore();
}

int ExtendedForm_getType(void* v)
{
	return ((CXairoIExtendedForm*) v)->getType();
}

char* ExtendedForm_getNorm(void* v)
{
	return utf8(((CXairoIExtendedForm*) v)->getNorm());
}

char* ExtendedForm_getText(void* v)
{
	return utf8(((CXairoIExtendedForm*) v)->getText());
}

char* ExtendedForm_getData(void* v,int i)
{
	return utf8(((CXairoIExtendedForm*) v)->getData(i));
}

void cleanup(int t,void* v)
{
	if (t==1) delete (CXairoICorpusTextFilter*) v;
	else if (t==2) delete (CXairoIWordList*) v;
	else if (t==3) delete (CXairoISolution*) v;
	else if (t==4) delete (CXairoIExtendedForm*) v;
}


xairo_uchar* uc(const char* s)
{
	int t=0;
	for (int i=0;s[i];i++) {
		char c=s[i];
		int cc=c & 0xE0;
		if (cc==0xE0) i+=2;
		else if (cc==0xC0) i++;
		t++;
	}
	xairo_uchar* w=new xairo_uchar[t+1];
	t=0;
	for (int i=0;s[i];i++) {
		char c=s[i];
		int cc=c & 0xE0;
		if (cc==0xE0) {
			w[t]=((s[i] & 0xf) << 12) +((s[i+1] & 0x3f) << 6) +(s[i+2] & 0x3f);
			i+=2;
		} 
		else if (cc==0xC0) {
			w[t]=((s[i] & 0xf) << 6) +(s[i+1] & 0x3f);
			i++;
		}
		else w[t]=s[i];
		t++;
	}
	w[t]=0;
	return w;
}




char* utf8(const xairo_uchar* w)
{
	int t=0;
	for (int i=0;w[i];i++) {
		if (w[i] <= 0x7f) t++;
		else if (w[i]<=0x7ff) t+=2;
		else t+=3;
	}
	char* s=new char[t+1];
	int j=0;
	for (int i=0;w[i];i++) {
		xairo_uchar x=w[i];
		if (x <= 0x7f) {
			s[j++]=(char) x;
		}
		else if (x <= 0x7ff){
			s[j++]=(char) (((x & 0x7C0)>>6) | 0xC0);
			s[j++]=(char) ((x & 0x3f)|0x80);
		}
		else {
			s[j++]=(char) (((x & 0xf000) >> 12) | 0xE0);
			s[j++]=(char) (((x & 0xFC0) >> 6) | 0x80);
			s[j++]=(char) ((x & 0x3F) |0x80);
		}
	}
	s[j]=0;
	return s;
}

char* _transcode(const char* c)
{
	return _strdup(utf8(transcode(uc(c))));
}

char* _transcode2(const char* c)
{
	return _strdup(utf8(transcode2(uc(c))));
}
